# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
# Copyright (c) 2005 Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

"""
Runtime checks.
"""

import gtk
import gobject
import gst
import sys
import os
import time

from configure import PYGTK_REQ, PYGST_REQ, GST_REQ

from codeina.extern.log import log

from gettext import gettext as _

OS_NAME = None
ARCH_NAME = None
DISTRO_NAME = None
GST_CORE_VERSION = None
GST_PLUGINS_BASE_VERSION = None

#############################################################################
# Version
#
# Brain-dead version class to easily parse and compare GStreamer-type
# version strings correctly (such as "0.10.10" > "0.10.0" etc.)
# Can probably do the same thing much easier with lists
#############################################################################


class Version:

    def __init__(self, version_string):
        self.major = 0
        self.minor = 0
        self.micro = 0
        self.nano = 0
        arr = version_string.split(".")
        if len(arr) > 0:
            self.major = int(arr[0])
            if len(arr) > 1:
                self.minor = int(arr[1])
                if len(arr) > 2:
                    self.micro = int(arr[2])
                    if len(arr) > 3:
                        self.nano = int(arr[3])

    def is_newer_than(self, other_version):
        if other_version.major != self.major:
            return other_version.major < self.major
        if other_version.minor != self.minor:
            return other_version.minor < self.minor
        if other_version.micro != self.micro:
            return other_version.micro < self.micro
        if other_version.nano != self.nano:
            return other_version.nano < self.nano
        return False  # is equal, so not newer

  # assuming self is a required version and available is the available version,
  # check if the available version fulfils the requirement

    def fulfils_requirement(self, available):

    #print "Checking %d.%d.%d.%d >= %d.%d.%d.%d" % (available.major,
    #    available.minor, available.micro, available.nano,
    #    self.major, self.minor, self.micro, self.nano)

        req = self
        if req.major != available.major:
            return req.major < available.major
        if req.minor != available.minor:
            return req.minor < available.minor
        if req.micro != available.micro:
            return req.micro < available.micro
        if req.nano != available.nano:
            return req.nano < available.nano
        return True

    def get_string(self):
        if self.nano:
            return "%d.%d.%d.%d" % (self.major, self.minor, self.micro,
                                    self.nano)
        else:
            return "%d.%d.%d" % (self.major, self.minor, self.micro)


#############################################################################
# get_plugin_version
#############################################################################


def get_plugin_version(plugin_name):
    plugin = gst.registry_get_default().find_plugin(plugin_name)
    if plugin:
        return Version(plugin.get_version())
    else:
        return None

#############################################################################
# get_plugin_from_element_factory_name
#############################################################################

def get_plugin_from_element_factory_name(name):
    registry = gst.registry_get_default()
    for plugin in registry.get_plugin_list():
        features = registry.get_feature_list_by_plugin (plugin.get_name())
        for f in features:
            if gobject.type_is_a(f.__gtype__, gst.ElementFactory.__gtype__):
                if f.get_name() == name:
                    return plugin
    return None
    
##############################################################################
# get_plugin_mtime
#
# returns time-module style tuple of the last modification time (UTC) of
# the plugin file of the given plugin object
#############################################################################

def get_plugin_mtime(plugin):
    if plugin and plugin.get_filename():
        s = os.stat(plugin.get_filename())
        if s:
            return time.gmtime(s.st_mtime)
    return None

##############################################################################
# get_element_mtime
#
# returns time-module style tuple of the last modification time (UTC) of
# the plugin file that contains the given element factory. Only works for
# external plugins.
#############################################################################

def get_element_mtime(name):
    plugin = get_plugin_from_element_factory_name(name)
    if plugin:
        return get_plugin_mtime(plugin)
    else:
        return None

#############################################################################
# get_element_version
#############################################################################


def get_element_version(element_name):
    feature = gst.registry_get_default().find_feature(element_name, gst.TYPE_ELEMENT_FACTORY)
    if feature:

    # hacking around missing GStreamer/gst-python API (how to get the plugin
    # or plugin_name from a GstPluginFeature?) This doesn't handle
    # nano-versions properly, ie. 0.10.10.1 will turn out to be 0.10.10, but
    # we don't really need that anyway

        for micro in range(199, -2, -1):
            if gst.default_registry_check_feature_version(element_name,
                    0, 10, micro):
                break
        if micro >= 0:

      #print "Element %s installed, Version: 0.10.%d" % (element_name, micro)

            return Version("0.10.%d" % micro)
        else:
            return None
    else:
        return None

#############################################################################
# _get_distro_id_from_lsb_release
#############################################################################

def _get_distro_id_from_lsb_release():
    try:
        s_in, s_out, s_err = os.popen3('lsb_release --id')
        output = s_out.read()
        if not s_out.close() and not s_err.close() and not s_in.close():
            if output:   # exit status 0 = all ok
                output = output.rstrip('\n\r\t ')
                output = output.lower()
                idx = max (output.rfind('\t'), output.rfind(':'))
                if idx >= 0:
                    return output[idx+1:]
                else:
                    return output
            else:
              return None
        else:
          return None
    except:
      return None

#############################################################################
# _get_distro_id_from_release_file
#############################################################################

def _get_distro_id_from_release_file():

# note: keep redhat-release towards the end, since mandrake has it too (sic)
# it's a tuple of tuples so we conserve the order
    mappings = (
        ('/etc/fedora-release',        'fedora'),
        ('/etc/mandrake-release',      'mandriva'),
        ('/etc/mandriva-release',      'mandriva'),
        ('/etc/mandrakelinux-release', 'mandriva'),
        ('/etc/SuSE-release',          'suse'),
        ('/etc/novell-release',        'suse'),
        ('/etc/redhat-release',        'redhat'),
        ('/etc/debian_version',        'debian'),
        ('/etc/netware-release',       'netware'),
        ('/etc/slackware-version',     'slackware'),
        ('/etc/yellowdog-release',     'yellowdog'),
    )

    for path, distro in mappings:
        if os.path.exists(path):
            return distro.lower()

    if OS_NAME:
        osname = OS_NAME.lower()
        if osname.find('sunos') >= 0 or osname.find('solaris') >= 0:
            return 'sunos'
        elif osname.find('freebsd') >= 0:
            return 'freebsd'
        elif osname.find('openbsd') >= 0:
            return 'openbsd'
        elif osname.find('netbsd') >= 0:
            return 'netbsd'

    return None

#############################################################################
# find_system_details
#############################################################################

def find_system_details():
    global OS_NAME
    global ARCH_NAME
    global DISTRO_NAME
    global GST_CORE_VERSION
    global GST_PLUGINS_BASE_VERSION

    u = os.uname()
    OS_NAME = u[0]
    ARCH_NAME = u[4]

  # for now munge everything x86 to i386

    if ARCH_NAME in ('i686', 'i586', 'i486', 'x86'):
        ARCH_NAME = "i386"

  # Find distro name
    DISTRO_NAME = _get_distro_id_from_lsb_release()

    if not DISTRO_NAME:
        DISTRO_NAME = _get_distro_id_from_release_file()

    if not DISTRO_NAME:
        DISTRO_NAME = 'generic'

  # Munge distro name to our liking (e.g. lsb-release will return 'fedoracore')
    if DISTRO_NAME.startswith('fedora'):
        DISTRO_NAME = 'fedora'

  # Find out current GStreamer core version

    GST_CORE_VERSION = Version(_version_to_string(gst.version()))

  # Find out current gst-plugins-base version

    GST_PLUGINS_BASE_VERSION = get_plugin_version('audioconvert')
    if not GST_PLUGINS_BASE_VERSION:
        GST_PLUGINS_BASE_VERSION = get_plugin_version('typefind')
    if not GST_PLUGINS_BASE_VERSION:
        GST_PLUGINS_BASE_VERSION = get_plugin_version('volume')
    if not GST_PLUGINS_BASE_VERSION:
        log.warning("check", "Could not determine gst-plugins-base version")
        GST_PLUGINS_BASE_VERSION = None
    log.info("check", "System Details:")
    log.info("check", "  OS  : %s" % OS_NAME)
    log.info("check", "  ARCH: %s" % ARCH_NAME)
    log.info("check", "  CORE: %s" % GST_CORE_VERSION.get_string())
    if GST_PLUGINS_BASE_VERSION:
        log.info("check", "  BASE: %s" % GST_PLUGINS_BASE_VERSION.get_string())
    else:
        log.info("check", "  BASE: ???")
    log.info("check", "DISTRO: %s" % DISTRO_NAME)

#############################################################################
# initial_checks
#############################################################################


def initial_checks():
    """
    do some basic check
    If there is a problem, will display the error and exit.
    """

    find_system_details()

    res = _checks()
    if not res:
        return
    (message, detail) = res

    # TODO check if we're running graphically

    dialog = gtk.MessageDialog(type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_OK)
    dialog.set_markup("<b>" + message + "</b>")
    dialog.format_secondary_text(detail)
    dialog.run()

    # 2 = GST_INSTALL_PLUGINS_ERROR

    sys.exit(2)


def __try_import__(modulename):
    try:
        __import__(modulename)
        return True
    except:
        return False


def _version_to_string(version):
    import string
    return string.join([str(x) for x in version], ".")


def _string_to_list(version):
    return [int(x) for x in version.split(".")]


def check_required_version(modulename):
    """
    Checks if the installed module is the required version or more recent.
    Returns [None, None] if it's recent enough, else will return a list
    containing the strings of the required version and the installed version.
    This function does not check for the existence of the given module !
    """

    if modulename == "gtk":
        if list(gtk.pygtk_version) < _string_to_list(PYGTK_REQ):
            return [PYGTK_REQ, _version_to_string(gtk.pygtk_version)]
    if modulename == "pygst":
        if list(gst.pygst_version) < _string_to_list(PYGST_REQ):
            return [PYGST_REQ, _version_to_string(gst.pygst_version)]
    if modulename == "gst":

        # use gst.version() here so we get the nano version too

        if list(gst.version()) < _string_to_list(GST_REQ):
            return [GST_REQ, _version_to_string(gst.version())]
        else:
            log.info("check",
                "GStreamer version ok (%s)" % _version_to_string(gst.version()))
    return [None, None]


def _checks():
    """
    Check versions of some required python modules.

    @returns: None on success, and a message and detail string on failure.
    @rtype:   None or tuple of (str, str)
    """
#    reg = gst.registry_get_default()
#    if not reg.find_plugin("autodetect"):
#        return (_("Could not find the autodetect plugins!"),
#                _("Make sure you have installed gst-plugins-good and is available in the GStreamer plugin path."))

    (req, inst) = check_required_version("gtk")
    if req:
        return (_("Version %s of the GTK+ Python Bindings is too old. ") % inst,
                _("Please install a newer version of the GTK+ Python bindings. "
                  "Version %s or greater is required to run this application.") % req)
    (req, inst) = check_required_version("pygst")
    if req:
        return (_("Version %s of the GStreamer Python bindings is too old.") % inst,
                _("Please install a newer version of the GStreamer Python bindings. "
                  "Version %s or greater is required to run this application..") % req)
    (req, inst) = check_required_version("gst")
    if req:
        return (_("Version %s of GStreamer is too old.") % inst,
                _("Please install a newer version of GStreamer. "
                  "Version %s or greater is required to run this application.") % req)
    return None
