# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
#
# Written by Tim-Philipp Müller <tim centricular net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import BaseHTTPServer
import gobject
import gtk
import os
import tempfile
import time
import thread
import urllib2
import xdg.BaseDirectory
import sha

from codeina.extern.log import log
from gettext import gettext as _

# Defaults to ~/.local/share/codeina/downloads/ in most cases
DEFAULT_DOWNLOAD_DIR = xdg.BaseDirectory.save_data_path('codeina/downloads/')

"""
HttpDownload

Downloads a HTTP URI to file in a separate thread and iterates the
GLib/Gtk main context until either the download is done or an error occurred
"""
class HttpDownload(log.Loggable):
    """
    @ivar  content_length: size in bytes of the download
    """

    CHUNKSIZE = 4096

    # small sleeps so the user has a chance to notice changes in the dialog
    # (throttles in effect, but shouldn't matter with our file sizes)
    # thomasvs: this was 0.1 originally, and this seriously made the
    # download slower - 8 seconds - instead of using full bandwidth, so
    # I changed it
    THROTTLE_SLEEP_TIME = 0.01

    def __init__ (self):
        self.url = None
        self.sha1sum = None
        self.directory = None
        self.fn = None
        self.done = False
        self.err = None
        self.bytes_read = -1
        self.content_length = 0
        self.cancelled = False
        self.verbose = False

    def get_error_msg_from_error(self, err):
        msg = _("Unexpected error of type %s") % type(err)
        try:
            raise err
        except urllib2.HTTPError, e:
            if e.code == 403:
              msg = _("Access to this file is forbidden")
            elif e.code == 404:
              msg = _("File not found on server")
            elif e.code == 500:
                msg = _("Internal server error, try again later")
            elif e.code == 503:
                msg = _("The server could not process the request due to " \
                        "high server load, try again later")
            else:
                try:
                    m = BaseHTTPServer.BaseHTTPRequestHandler.responses[e.code]
                except KeyError:
                    m = _("Unknown Error")
                msg = _("Server error %u: %s") % (e.code, m)
        except urllib2.URLError, e:
            msg = _("Failed to connect to server: %s") % e.reason[1]
        except IOError, e:
            if e.filename and e.strerror:
                msg = _("File error for '%s': %s") % (e.filename, e.strerror)
            elif e.strerror:
                msg = _("File error: %s") % e.strerror
            elif e.filename:
                msg = _("File error for '%s'") % e.filename
            elif e.errno:
                msg = _("Unknown file error %d") % e.errno
            else:
                msg = _("Unknown file error %d") % e.errno

        return msg

    """ Iterates the default GLib main context until the download is done """
    def run(self, url, directory, fn, sha1sum, message=None):
        self.debug("Running download for %s (sha1sum=%s)" % (os.path.join(directory, fn), sha1sum))
        if sha1sum and self.verify(os.path.join(directory, fn), sha1sum):
            self.debug("Already downloaded %s" % os.path.join(directory, fn))
            return None

        if self.url:
            return _("Download already in progress")

        if message:
            print message

        self.url = url
        self.directory = directory
        self.fn = fn
        self.sha1sum = sha1sum
        self.done = False
        self.err = None
        self.bytes_read = -1
        self.content_length = 0
        self.cancelled = False

        thread.start_new_thread(self.thread_func, ())

        timeout_id = gobject.timeout_add(500, self.update_progress)

        self.update_progress()

      # Don't try to be smart and block here using mayblock=True, it won't work
        while not self.done:
            gobject.main_context_default().iteration(False)

        gobject.source_remove(timeout_id)

      # Update to 'finished' state
        self.update_progress()
        while gobject.main_context_default().pending():
            gobject.main_context_default().iteration(False)

      # Reset state
        self.url = None
        self.directory = None
        self.fn = None
        self.done = False
        self.bytes_read = -1
        self.content_length = 0

        return self.err

    """ Called from the context of the main GUI thread """
    def update_progress(self):
        if self.verbose:
            if self.bytes_read < 0:
                print "Connecting to server"
            elif self.bytes_read == 0:
                print "Connected to server"
            elif self.content_length > 0:
                print "Bytes read: %.1f (%.1f%%)" %                    \
                      (self.bytes_read / 1024.0,                       \
                       self.bytes_read * 100.0 / self.content_length)
            else:
                print "Bytes read: %.1f" % (self.bytes_read / 1024.0)

        return not self.done

    """ Check whether a file matches the SHA1 sum given """
    def verify(self, fn, sha1sum):
        if not sha1sum:
            self.debug("No sha1sum or filename, can't check (%s), "
                "so assuming true" % fn)
            return True

        fs = open(fn,'rb').read()
        real_sha1sum = sha.new(fs).hexdigest()
        if real_sha1sum == sha1sum:
            self.debug("%s matched its sha1sum" % fn)
            return True

        self.debug("%s didn't match its sha1sum (got %s expected %s)" % (
            fn, real_sha1sum, sha1sum))
        return False

    """ Thread where all the downloading and file writing happens """
    def thread_func(self):

        tmpfile_path = None

      # First, see if we can resolve the URL at all and connect to the host
        try:
            self.bytes_read = -1
            res = urllib2.urlopen(self.url)
            self.bytes_read = 0
            time.sleep(self.THROTTLE_SLEEP_TIME)

            try:
                self.content_length = int(res.info()['Content-Length'])
            except KeyError:
                self.content_length = 0

          # Make sure the directory exists (ignore exception thrown if it
          # already exists, mkstemp will throw an exception later if there
          # is really a problem)
            try:
                os.makedirs(self.directory)
            except:
                pass

          # Now create a temp file there
            tmpfile_handle, tmpfile_path = tempfile.mkstemp('.incomplete',
                                                            'codeina-download',
                                                            self.directory)

            while not self.cancelled:
                data = res.read(self.CHUNKSIZE)
                if data and len(data):
                    self.bytes_read += len(data)
                    os.write(tmpfile_handle, data)
                    # give GUI thread a chance to show progress
                    time.sleep(self.THROTTLE_SLEEP_TIME)
                else:
                    break

          # Now rename the temporary file to the desired file name
            if self.cancelled:
                try:
                    os.remove(tmpfile_path)
                except:
                    pass
            elif not self.verify(tmpfile_path, self.sha1sum):
                try:
                    os.remove(tmpfile_path)
                except:
                    pass
                err = _("Downloaded file is corrupted")
            else:
                os.rename(tmpfile_path, os.path.join(self.directory, self.fn))

        except (urllib2.HTTPError, urllib2.URLError, IOError), e:
            self.err = self.get_error_msg_from_error (e)
            if tmpfile_path:
                try:
                    os.remove(tmpfile_path)
                except:
                    pass
        else:
            self.err = None

      # Finished (one way or another), stop the thread
        self.done = True
        thread.exit()

"""
HttpDownloadDialog

Same as HttpDownload, but shows a progress dialog.
Note that it doesn't derive from GtkDialog.
"""
class HttpDownloadDialog(HttpDownload):
    def __init__(self, parent_window=None):
        HttpDownload.__init__(self)
        self.parent_window = parent_window
        self.create_dialog()

    def dialog_response_cb(self, *args):
        if not self.done:
            dlg, response_id = args
            if response_id in (gtk.RESPONSE_REJECT, gtk.RESPONSE_DELETE_EVENT):
                self.cancelled = True
            else:
                print "Unexpected download dialog response id", response_id

    def create_dialog(self, message=None, parent_window=None):
        self.dlg = gtk.Dialog(title=_("Downloading"),\
                              parent=parent_window,  \
                              buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT))
        alignment = gtk.Alignment(0.0, 0.5, 1.0, 1.0)
        alignment.set_padding (6, 6, 6, 6)
        self.dlg_label = gtk.Label()
        if message:
            self.set_dialog_text(message)
        alignment.add(self.dlg_label)
        self.dlg.vbox.pack_start(alignment)
        self.progressbar = gtk.ProgressBar()
        self.dlg.vbox.pack_start(self.progressbar)
        self.dlg.connect('response', self.dialog_response_cb)
        self.dlg.show_all()
        self.dlg.hide()

    def set_dialog_text(self, txt):
        self.dlg_label.set_markup('<b>' + gobject.markup_escape_text(txt) + '</b>')

    def show_dialog(self):
        self.dlg.show()

    def hide_dialog(self):
        self.dlg.hide()

    def destroy(self):
        self.dlg.destroy()

    def run(self, url, directory, fn, sha1sum, message):
        if message:
            self.set_dialog_text(message)
        self.dlg.present()
        err = HttpDownload.run(self, url, directory, fn, sha1sum, None)
        self.dlg.hide()
        return err

    """ Called from the context of the main GUI thread """
    def update_progress(self):
        if self.done and not self.err and not self.cancelled:
            self.progressbar.set_text(_("Finished"))
            self.progressbar.set_fraction(1.0)
        elif self.done and (self.err or self.cancelled):
            self.progressbar.set_text('')
        elif self.bytes_read < 0:
            self.progressbar.set_text(_("Connecting to server"))
            self.progressbar.set_fraction(0.0)
        elif self.bytes_read == 0:
            self.progressbar.set_text(_("Connected to server"))
            self.progressbar.set_fraction(0.0)
        elif self.content_length > 0:
            self.progressbar.set_text(_("Downloading (%.1f KB of %.1f KB)") % (
                self.bytes_read / 1000.0, self.content_length / 1000.0))
            percent = (float) (self.bytes_read * 1.0 / self.content_length)
            self.progressbar.set_fraction(percent)
        else:
            self.progressbar.set_text(_("Downloading"))
            self.progressbar.pulse()
        return not self.done


if __name__ == "__main__":
    log.init('CODEINA_DEBUG')
    log.setPackageScrubList('codeina')

    # Different URLs for different errors
    #url='http://shop.fluendo.com/images/mpeg.png'
    #url='http://shop.fluendo.com/tpm-codeina-test.bz2'
    #url='http://example.com/gimme-a-404'
    #url='http://antihost.atari/tpm-codeina-test.bz2'
    url="http://shop.fluendo.com/pub/fluendo-mp3-2.i386.tar.bz2"
    sha1sum='117f85e118dab2d9f7c9bb10554fbc4e8f7fdf38'

    download_dir = DEFAULT_DOWNLOAD_DIR

    fn = os.path.basename(url)

  # Do the downloading, with dialog
    d = HttpDownloadDialog()
    e = d.run(url, download_dir, fn, sha1sum, "Downloading '%s'" % fn)
    if d.cancelled:
        print "Download cancelled"
    elif e:
        print "Download error:", e
    else:
        print "Download done, no error."
