# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
#
# Written by Tim-Philipp Müller <tim centricular net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
Install a .tar.gz file containing GStreamer plugins into the user's home
directory (~/.gstreamer-0.10/plugins/)
"""

import os
import gst
import myutils
import tarfile
import tempfile
import sys

from codeina.extern import which
from codeina.extern.log import log

"""
_plugin_file_can_be_loaded
"""
def _plugin_file_can_be_loaded(plugin_file_path):
    try:
      child = os.popen("gst-inspect-0.10 '%s'" % plugin_file_path)
      child.read()
      if not child.close():              # exit status 0 = all ok
        return True
      else:
        return False
    except:
      return False

"""
install_plugin_archive:

installs the given .tar.gz (or .tar.bz2) plugin archive into
the user's ~/.gstreamer-0.10/plugins/ directory. Returns list of plugin names
which failed to install/load, or None/an empty list if all went fine.
"""
def install_plugin_archive(archive_path):
    tar = tarfile.open(archive_path, 'r')

    basename = os.path.basename(archive_path)
    if basename.endswith('.tar.gz'):
        basename = basename[0:-7]
    elif basename.endswith('.tar.bz2'):
        basename = basename[0:-8]

# make sure ~/.gstreamer-0.10 and ~/.gstreamer-0.10/plugins exist
    user_gst_dir = os.path.join(os.environ['HOME'],'.gstreamer-0.10')
    user_plugins_dir = os.path.join(user_gst_dir,'plugins')

    try:
      os.makedirs(user_plugins_dir)
    except OSError:
      pass

# create temporary directory under ~/.gstreamer-0.10

    tdir = tempfile.mkdtemp('', basename + '.dir.', user_gst_dir)

# extract each plugin file in the tar archive to a temp folder, and move
# it over into ~/.gstreamer-0.10/plugins after checking that we can in fact
# load it without errors

    error_names = [] 
    for name in tar.getnames():
        if name.endswith('.so') \
        or name.endswith('.sl') \
        or name.endswith ('.dll') \
        or name.endswith ('.DLL') \
        or name.endswith ('.dynlib'):
            tar.extract(name, tdir)
            plugin_tmp_path = os.path.join(tdir, name)
            plugin_dest_path = os.path.join(user_plugins_dir,
                                            os.path.basename(name))
            # thomas: UGLY HACK: on Fedora, we need to allow text relocations
            # for the i386 IPP mp3 plugin
            # in the future, do this more generically, perhaps with an add-on
            # script in the tarball itself ?
            if which.which('chcon') and 'i386' in basename and 'mp3' in name:
                log.debug('install', "modifying mp3 plug-in's secontext")
                status = os.system('chcon --type=textrel_shlib_t %s' % 
                    plugin_tmp_path)
                if status:
                    log.warning('install',
                        'Failed to change secontext, status %r' %
                            status)
            
            if _plugin_file_can_be_loaded(plugin_tmp_path):
                os.system("mv '%s' '%s'" % (plugin_tmp_path, plugin_dest_path))
                if not gst.plugin_load_file(plugin_dest_path):
                    log.warning('codeina', "Problem loading plugin %s", 
                        plugin_dest_path)
                    error_names.append(name)
                log.debug('install', "plugin %s loaded successfully",
                    plugin_dest_path)
            else:
                log.log('codeina', "Problem introspecting plugin %s", name)
                error_names.append(name)

# close tar file and remove temporary directory

    tar.close()
    os.system("rm -rf '%s'" % tdir)

    return error_names


# if program is called directly, install the given archive file(s)

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print "Usage: %s [PLUGINS_BUNDLE.TAR.GZ|PLUGINS_BUNDLE.TAR.BZ2]" % \
              (sys.argv)[0]
        sys.exit(1)

    for archive in (sys.argv)[1:]:
        basename = os.path.basename(archive)

        print "Installing:", basename
        try:
            error_names = install_plugin_archive(archive)
            if error_names:
                print "Failed to install these plugins from %s:" % archive
                print error_names
            else:
                print "OK"
        except Exception, e:
            print "Failed to install %s: %s" % (basename, myutils.get_exception_message(e))
        print ''

    sys.exit(0)

