# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
#
# Written by Tim-Philipp Müller <tim centricular net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import pygtk
pygtk.require('2.0')
import gtk
import pango

from productlist import Product

from gettext import gettext as _

# List widget rendering a product list


class ListView(gtk.VBox):
    """
    @ivar productlist: list of products, set when creating the listview
    @type productlist: L{codeina.productlist.ProductList}
    """

    def __init__(self):
        gtk.VBox.__init__(self)
        self.required_plugins = []

      # list view label

        self.listlabel = gtk.Label()
        self.listlabel.set_alignment(0.01, 0.5)
        self.listlabel.set_padding(0, 3)
        self.listlabel.show()
        self.pack_start(self.listlabel, False)

      # list view

        self.view = gtk.TreeView()
        row_height = self.get_estimated_row_height()
        row_padding = 2  # assumed, random number
        headers_height = 12 # assumed
        self.view.set_size_request(-1,
            (row_height + row_padding) * 8 + headers_height)
        self.productlist = None
        self.store = gtk.ListStore(object)
        self.store.set_default_sort_func(self.sort_func, self)
        self.store.set_sort_column_id(-1, gtk.SORT_ASCENDING)
        self.model = self.store.filter_new()
        self.model.set_visible_func(self.visible_func, self)
        self.view.set_model(self.model)
        self.view.set_rules_hint(True)
        column = gtk.TreeViewColumn(_("Product"))
        text_renderer = gtk.CellRendererText()
        text_renderer.set_property('xpad', 6)
        text_renderer.set_property('ypad', 3)
        text_renderer.set_property('width-chars', 65)
        text_renderer.set_property('ellipsize', pango.ELLIPSIZE_END)
        column.pack_start(text_renderer, True)
        column.set_cell_data_func(text_renderer, self.cell_data_func_text,
                                  self)
        column.set_expand(True)
        self.view.append_column(column)

      # Add separate columns for install-status and price so we can set
      # the expand property to False, so that things behave right if the
      # user makes the window wider

        column = gtk.TreeViewColumn(_('Installed?'))
        status_renderer = gtk.CellRendererText()
        column.pack_start(status_renderer, True)
        status_renderer.set_property('xalign', 0.5)
        status_renderer.set_property('xpad', 12)
        status_renderer.set_property('text', 'not installed')
        column.set_cell_data_func(status_renderer, self.cell_data_func_status,
                                  self)
        column.set_expand(False)
        self.view.append_column(column)

        column = gtk.TreeViewColumn(_('Price'))
        price_renderer = gtk.CellRendererText()
        column.pack_end(price_renderer, False)
        price_renderer.set_property('xalign', 1.0)
        price_renderer.set_property('xpad', 6)
        column.set_cell_data_func(price_renderer, self.cell_data_func_price,
                                  self)
        column.set_expand(False)
        self.view.append_column(column)

     # Put check boxes in separate column for nicer activation

        column = gtk.TreeViewColumn("")
        toggle_renderer = gtk.CellRendererToggle()
        toggle_renderer.set_property('activatable', True)
        toggle_renderer.connect('toggled', self.cell_toggled_cb, self)
        toggle_renderer.set_property('xalign', 0.5)
        toggle_renderer.set_property('xpad', 6)
        column.pack_end(toggle_renderer, False)
        column.set_cell_data_func(toggle_renderer, self.cell_data_func_toggle,
                                  self)
        column.set_expand(False)
        self.view.insert_column(column, 0)

      # pack tree view in scrolled window

        scrollwin = gtk.ScrolledWindow()
        scrollwin.show()
        scrollwin.add(self.view)
        scrollwin.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        scrollwin.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.pack_start(scrollwin, True)
        self.show_all()
        return

    def sort_func(self, tree_model, iter1, iter2, user_data):
        product1 = tree_model.get_value(iter1, 0)
        product2 = tree_model.get_value(iter2, 0)
        if product1 and not product2:
            return -1
        elif product2 and not product1:
            return 1
        elif product2.rank != product1.rank:
            return product2.rank - product1.rank
        elif product2.price != product1.price:
            return product2.price - product1.price
        else:
            return cmp(product1.name, product2.name)

    def visible_func(self, tree_model, tree_iter, user_data):
        product = tree_model.get_value(tree_iter, 0)
        if not product:
            return False
        flav = product.get_best_flavor_for_current_system()
        if not flav:
            product.set_selected(False)
            return False
        if not self.required_plugins:
            return True
        else:
            for req in self.required_plugins:
                if product.provides_requirement(req):
                    return True
            return False

    def cell_toggled_cb(self, cell, path, user_data):
        tree_iter = self.model.get_iter(path)
        product = self.model.get_value(tree_iter, 0)
        product.set_selected(not product.get_selected())
        self.model.row_changed(path, tree_iter)

    def cell_data_func_toggle(self, col, renderer, tree_model, tree_iter,
                              data):
        product = tree_model.get_value(tree_iter, 0)
        renderer.set_active(product.get_selected())
        if product.is_fully_installed() and not product.updates_available():
            renderer.set_property('visible', False)
        else:
            renderer.set_property('visible', True)

    def get_product_markup(self, product):
        markup = "%s\n<span size='smaller' weight='light'>%s</span>" % (product.name,
                product.summary)
        return markup

    def cell_data_func_text(self, col, renderer, tree_model, tree_iter,
                            data):
        product = tree_model.get_value(tree_iter, 0)
        markup = self.get_product_markup(product)
        renderer.set_property('markup', markup)
        return

    def get_estimated_row_height(self):
        layout = self.view.create_pango_layout('')
        fake_product = Product(**{'name': 'Whatevery', 'short-description': 'Woygg'})
        layout.set_markup(self.get_product_markup(fake_product))
        (w, h) = layout.get_pixel_size()
        return h

    def cell_data_func_status(self, col, renderer, tree_model, tree_iter,
                              data):
        product = tree_model.get_value(tree_iter, 0)

    # check status

        updates_available = product.updates_available()
        fully_installed = product.is_fully_installed()
        part_installed = product.is_partially_installed()

    # render status
    # if a package is partially installed, this can mean one of two things:
    #  - it is a bundle product and one of the things contained in the
    #        bundle is installed, but not the bundle itself
    #  - a product is available in a newer version that contains more elements
    #        than the currently installes version

        txt = ''
        if not fully_installed and not part_installed:
            txt = _('not installed')
        elif updates_available:
            txt = _('updates available')
        elif fully_installed:
            txt = _('installed')
        elif part_installed:
            txt = _('partially installed')
        else:
            txt = 'unknown/fixme'  # shouldn't happen

    # we assume txt won't contain any markup characters that would need escaping

        renderer.set_property('markup',
                              "<span size='smaller' weight='light'>%s</span>" %
                              txt)
        return

    def cell_data_func_price(self, col, renderer, tree_model, tree_iter,
                             data):
        product = tree_model.get_value(tree_iter, 0)
        if product.price:
            renderer.set_property('text', "%.0f €" % product.price)
        else:
            renderer.set_property('text', _("Free"))
        return

    def fill_with_products(self):
        self.store.clear()
        if self.productlist:
            for p in self.productlist.products:
                row = self.store.append()
                self.store.set_value(row, 0, p)
        return

    def set_product_list(self, product_list):
        """
        @ivar productlist: list of products the view should handle
        @type productlist: L{codeina.productlist.ProductList}
        """
        if self.productlist != product_list:
            self.productlist = product_list
            self.fill_with_products()
        return

    def show_all_available(self):
        self.listlabel.set_markup('<b>' + _("Available products") +
                                  ':</b>')
        self.required_plugins = None
        self.model.refilter()
        return

    def show_matching(self, required_plugins):
        if required_plugins == None or len(required_plugins) == 0:
            self.show_all_available()
            return

        self.listlabel.set_markup('<b>' + _("Matching products") +
                                  ':</b>')
        self.required_plugins = required_plugins
        self.model.refilter()
        return

    def get_visible_rows(self):
        return self.model.iter_n_children(None)


