# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
#
# Written by Tim-Philipp Müller <tim centricular net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
Main application
"""

import gobject
import gtk

import os
import sys
import traceback

from codeina import configure, check, install, myutils, productlist
from codeina import euladialog
from codeina.listview import ListView
from codeina.requiredplugin import RequiredPlugin
from codeina.httpdownload import HttpDownloadDialog, DEFAULT_DOWNLOAD_DIR
from codeina.extern.log import log

from gettext import gettext as _
from gettext import ngettext

# exit code constants

GST_INSTALL_PLUGINS_SUCCESS = 0
GST_INSTALL_PLUGINS_NOT_FOUND = 1
GST_INSTALL_PLUGINS_ERROR = 2
GST_INSTALL_PLUGINS_PARTIAL_SUCCESS = 3
GST_INSTALL_PLUGINS_USER_ABORT = 4

# distro dialogue constants

GST_DISTRO_DIALOG_ABORT = 0
GST_DISTRO_DIALOG_CONTINUE = 1

class InfoBox(gtk.HBox):

    def __init__(self):
        gtk.HBox.__init__(self)
        self.set_spacing(12)
        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_DIALOG_INFO, gtk.ICON_SIZE_DIALOG)
        img.show()
        self.pack_start(img, False)
        self.label_alignment = gtk.Alignment()
        self.label_vbox = gtk.VBox()
        self.primary_label = gtk.Label()
        self.primary_label.set_alignment(0.0, 0.5)
        self.primary_label.show()
        self.label_vbox.pack_start(self.primary_label, False)
        self.secondary_label = gtk.Label()
        self.secondary_label.set_alignment(0.0, 1.0)
        self.secondary_label.set_padding(0, 3)
        self.secondary_label.show()
        self.label_vbox.pack_start(self.secondary_label, False)
        self.label_alignment.add(self.label_vbox)
        self.pack_start(self.label_alignment, False)
        self.label_alignment.set(0.0, 0.5, 1.0, 1.0)
        self.show_all()
        self.show_default()
        self.set_border_width(12)

    def show_default(self):
        self.primary_label.set_text("No info")  ### FIXME ###
        self.secondary_label.set_text('')
        self.hide()

    def show_required_plugins(self, required_plugins):
        if required_plugins == None or len(required_plugins) == 0:
            self.show_default()
            return

        execname = required_plugins[0].application
        appname = myutils.get_application_name_from_execname(execname)
        txt = '<b>'
        txt += ngettext(
            "The application \"%s\" requires the following plugin:",
            "The application \"%s\" requires the following plugins:",
                len(required_plugins)) % appname
        txt += '</b>'
        self.primary_label.set_markup(txt)
        txt = ''
        first = True
        for req in required_plugins:
            #self.debug("REQUIRED: %s-%s", req.req_type, req.req_detail)
            if not first:
                txt += '\n'
            txt += req.description
            first = False
        self.secondary_label.set_text(txt)
        self.show()


# Main application class and application window


class CodeinaApp(gtk.Window, log.Loggable):

    ui_desc = \
        """<ui>
    <menubar name="MenuBar">
      <menu action="File">
        <menuitem action="FileInstallArchives"/>
        <separator />
        <menuitem action="FileQuit"/>
      </menu>
      <menu action="View">
        <menuitem action="ViewAllAvailable"/>
        <menuitem action="ViewMatching"/>
      </menu>
      <menu action="Help">
        <menuitem action="HelpAbout"/>
      </menu>
    </menubar>
    </ui>"""

    VIEW_ALL = 1
    VIEW_MATCHING = 2

    def __init__(self, *args):
        gtk.Window.__init__(self)

        self.debug('Creating window')
        # FIXME: we should load product list later (from http)

        self.parser = productlist.ProductListParser()
        try:
            xml_data = open(configure.make_xml_path ('available-plugins.xml'))
            self.product_list = self.parser.parse(xml_data)
        except:
            print "Oops, couldn't load products XML list:", sys.exc_info()[0]
            traceback.print_exc()

      # set default exit code

        self.exit_code = GST_INSTALL_PLUGINS_USER_ABORT
        self.installed_some = False

      # parse command line arguments

        self.required_plugins = []
        for arg in args[0]:
            if arg[0:16] == "--transient-for=":
                print "XID:", arg[16:]
            if arg[0:15] == "gstreamer|0.10|":
                req = RequiredPlugin(arg)
                if req.valid:
                    self.required_plugins.append(req)
                    self.debug("INSTALL PLUGIN: %r", arg[15:])

      # set up UIManager

        self.uimanager = gtk.UIManager()
        self.add_accel_group(self.uimanager.get_accel_group())
        self.actiongroup = gtk.ActionGroup("codeina actions")
        self.actiongroup.add_actions([
            ('File', None, _("_File"), None),
            ('FileInstallArchives', gtk.STOCK_OPEN,
                _('Install downloaded plugin archive'), None,
                 'Install downloaded .tar.gz or .tar.bz2 plugin archives',
                self.action_install_archive),
            ('FileQuit', gtk.STOCK_QUIT, None, None,
                _('Quit this application'), self.action_quit),
            ('View', None, _("_View"), None),
            ('Help', None, 'menu-action-dummy-2', None, None),
            ('HelpAbout', gtk.STOCK_ABOUT,
                None, None, _('About this application'), self.action_about)])
        self.actiongroup.add_radio_actions([
            ('ViewAllAvailable', None, _('_All available'),
                None, _('Show all available plugin packages'), self.VIEW_ALL),
            ('ViewMatching', None, _('_Matching'), None,
                'Show only those plugin packages matching the required ones',
                self.VIEW_MATCHING)], self.VIEW_ALL, self.action_toggle_view)

      # Get translated label from Gtk+ (doesn't work for STOCK_FILE sadly)

        item = gtk.stock_lookup(gtk.STOCK_HELP)
        self.actiongroup.get_action('Help').set_property('label', item[1])
        self.uimanager.insert_action_group(self.actiongroup, 0)
        self.uimanager.add_ui_from_string(self.ui_desc)

      # set up GUI bits

        self.set_title(_("Codec Installer"))
        self.connect("delete-event", self.delete_event)
        self.vbox = gtk.VBox()
        self.add(self.vbox)
        self.vbox.show()
        #self.show()

      # menu

        menubar = self.uimanager.get_widget('/MenuBar')
        self.vbox.pack_start(menubar, False)
        menubar.show()

      # info box

        self.infobox = InfoBox()
        self.infobox.show_required_plugins(self.required_plugins)
        self.vbox.pack_start(self.infobox, False)

      # create get button before we let the list view update sensitivity

        button = self._make_button_with_icon(gtk.STOCK_OK, _('_Get selected'))
        self.get_button = button
        self.get_button.set_sensitive(False)
        self.get_button.connect('clicked', self.get_selected_button_clicked)

      # list view

        self.listview = ListView()
        self.listview.show()
        self.listview.set_product_list(self.product_list)
        self.listview.show_all_available()
        self.listview.model.connect('row-changed',
            lambda t, p, i: self.update_button_sensitivity())

        self.vbox.pack_start(self.listview, True)

      # The label showing we have no matching plugins

        self.no_plugins_box = gtk.HBox()
        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_DIALOG_WARNING, gtk.ICON_SIZE_DIALOG)
        self.no_plugins_box.pack_start(img, False, 0, 6)
        label = gtk.Label(
            _("There are no plugins available for the requested functionality.")
            + "\n" + _("Please select the full plugins list in the View menu."))
        self.no_plugins_box.pack_start(label, False)
        self.vbox.pack_start(self.no_plugins_box, False)

      # status bar

        self.statusbar = gtk.Statusbar()
        self.statusbar_ctx_id = self.statusbar.get_context_id('default')
        self.statusbar.show()
        self.vbox.pack_end(self.statusbar, False)

      # only show matching plugin packages if there are required plugins

        if self.required_plugins and len(self.required_plugins) > 0:
            self.debug("Showing the matching plugins list")
            self.actiongroup.get_action('ViewMatching').set_active(True)
        else:
            self.debug("Showing all the plugins")
            self.actiongroup.get_action('ViewAllAvailable').set_active(True)
            self.actiongroup.get_action('ViewMatching').set_sensitive(False)

      # button box

        self.button_box = gtk.HButtonBox()
        self.button_box.show()

      # close button

        button = gtk.Button(stock=gtk.STOCK_CLOSE)
        button.connect('clicked', self.close_button_clicked)
        button.show()
        self.button_box.add(button)

      # get button

        self.button_box.add(self.get_button)

      # this positions the get button at the right (in a LTR locale)

        self.button_box.set_child_secondary(self.get_button, True)
        self.button_box.set_border_width(6)
        self.vbox.pack_start(self.button_box, False)

        for req in self.required_plugins:
            for p in self.product_list.products:
                if p.provides_requirement(req):
                    self.debug(" - provided by product '%s'" % p.name)

        if len(self.product_list.products) == 0:
            self.set_status_text('Failed to load available products list')

    def show_distro_dialog(self):
        self.debug('show_distro_dialog(): start')
        # show disclaimer dialog
        dismiss_path = os.path.join(os.path.expanduser('~'), ".gnome2", 
            "codeina_dismissed")
        if os.path.exists(dismiss_path):
            return GST_DISTRO_DIALOG_CONTINUE
        try:
            module = "codeina.distro.%s" % check.DISTRO_NAME
            distro = __import__(module, globals(), locals(),
                ["codeina", "distro"])
            dialog = distro.get_dialog(self, self.required_plugins)
            if dialog:
                dialog.set_transient_for(self)
                button = gtk.CheckButton(
                    label=_("Do not show me this message _again"))
                dialog.vbox.add(button)
                button.show()
                self.debug('show_distro_dialog(): running dialog')
                resp = dialog.run()
                if button.get_active():
                    dir_path = os.path.join(os.path.expanduser('~'),
                        ".gnome2")
                    if not os.path.exists(dir_path):
                        os.mkdir(dir_path)
                    dismiss_file = open(dismiss_path, 'w')
                    dismiss_file.close()
                if resp != gtk.RESPONSE_OK:
                    self.debug('show_distro_dialog(): abort')
                    return GST_DISTRO_DIALOG_ABORT
                dialog.destroy()
        except ImportError:
            # no distro specific dialog, so don't show anything
            pass

        self.debug('show_distro_dialog(): ok')
        return GST_DISTRO_DIALOG_CONTINUE

    def close_button_clicked(self, user_data):
        self.shutdown()

    def _make_button_with_icon(self, stock_id, custom_txt):
        bhbox = gtk.HBox()
        bimg = gtk.Image()
        bimg.set_from_stock(stock_id, gtk.ICON_SIZE_BUTTON)
        bhbox.pack_start(bimg, False, 0, 6)
        blabel = gtk.Label()
        blabel.set_text_with_mnemonic(custom_txt)
        bhbox.pack_start(blabel, False)
        button = gtk.Button()
        button.add(bhbox)
        button.show_all()
        button.set_property('can-default', True)
        return button

    def download_uri(self, dialog, name, uri, sha1sum):
        """
        Download the given uri.

        @param dialog: L{codeina.httpdownload.HttpDownloadDialog}

        @returns: path to the downloaded file, or None.
        """
        self.debug("Trying to download '%s'" % uri)
        dl_dir = DEFAULT_DOWNLOAD_DIR
        fn = os.path.basename(uri)
        err_msg = dialog.run(uri, dl_dir, fn, sha1sum,
            _("Downloading '%s'") % name)
        if dialog.cancelled:
            self.debug("Download cancelled")
            return None
        elif err_msg:
            dlg = gtk.MessageDialog(parent=self, flags=0,
                                    type=gtk.MESSAGE_ERROR,
                                    message_format=_('Download failed'))
            dlg.format_secondary_text(
                _("Failed to download '%s' from \n%s\n"
                  "Reason: %s") % (name, uri, err_msg))
            dlg.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
            dlg.set_default_response(gtk.RESPONSE_CLOSE)
            dlg.run()
            dlg.destroy()
            return None

        self.debug("Downloaded '%s'" % fn)
        return os.path.join(dl_dir, fn)

    def download_and_show_eula(self, dialog, name, eula_uri):
        """
        Shows EULA and blocks.

        Returns: whether the EULA was accepted.
        """
        path = self.download_uri(dialog, name, eula_uri, None)
        if not path:
            self.debug('Could not download EULA')
            return False
    
        # run a dialog showing the eula
        eula = open(path).read()
        d = euladialog.EulaDialog(self, eula)
        result = d.run()
        d.destroy()
        if result == gtk.RESPONSE_OK:
            self.debug('User accepted EULA')
            return True

        self.debug('User refused EULA')
        return False

    def download_free_products(self, products):
        """
        Returns: whether the user downloaded all products.
        """
        # we assume that for free products the download location is public and
        # does not require authentication
        # We may have to show a EULA however.
        d = HttpDownloadDialog(parent_window=self)
        files_downloaded = []
        for p in products:
            self.debug('Downloading product %s', p.name)
            if p.eula_uri:
                self.debug('Downloading EULA %s', p.eula_uri)
                if not self.download_and_show_eula(d, p.name, p.eula_uri):
                    continue

            flavor = p.get_best_flavor_for_current_system()
            self.debug("Selected flavour %s %s %s %s %s" % (
                flavor.system, flavor.arch, flavor.distro,
                os.path.basename(flavor.uri), flavor.sha1sum))
            path = self.download_uri(d, p.name, flavor.uri, flavor.sha1sum)

            if not path:
                self.debug('User cancelled')
                return False

            if path:
                self.debug("Downloaded '%s'" % path)
                files_downloaded.append(path)

        # If we downloaded any packages, try to install them now
        self.debug("%d files downloaded", len(files_downloaded))
        if files_downloaded:
            d.progressbar.set_text('')
            d.set_dialog_text(_("Installing downloaded packages"))
            d.show_dialog()
            while gobject.main_context_default().pending():
                gobject.main_context_default().iteration(False)

            self.do_install(files_downloaded)
            d.destroy()

        return True

    def get_products(self, products):
      # TODO:
      #   (a) check if any of the remaining selected products are part of
      #       another selected product (see product.is_part_product_of(p2))
      #
      #   (b) show disclaimer/message dialog on restricted formats

        for p in products:
            self.debug("GET '%s'" % p.name)

        dlg = gtk.MessageDialog(parent=self, flags=0,
                                type=gtk.MESSAGE_INFO,
                                message_format=_('Getting plugins'))
        dlg.set_title(_('Getting plugins'))
        txt = _('At the moment this application does not support '
                'purchasing plugins directly from the Fluendo web shop '
                'yet.')
        txt += '\n\n'
        txt += _('Please buy and download the selected plugins from the '
                 'web shop and once you have done that install them '
                 'using the \'Install downloaded plugin archive\' menu '
                 'item from the File menu.')
        txt += '\n\n'
        txt += _('I will now open the Fluendo web shop in a browser.')
        dlg.format_secondary_text(txt)
        dlg.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        button = self._make_button_with_icon(gtk.STOCK_OK,
            _('Start Web Browser'))
        dlg.add_action_widget(button, gtk.RESPONSE_OK)
        dlg.set_default_response(gtk.RESPONSE_OK)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            u = myutils.append_system_details_to_url('http://shop.fluendo.com')
            myutils.open_url_in_browser(u)
        dlg.destroy()
        return

    def get_selected_button_clicked(self, user_data):
        self.debug('User clicked get selected')
        free_products = []
        pay_products = []
        for p in self.product_list.get_selected():
            if p.price == 0:
                free_products.append(p)
            else:
                pay_products.append(p)
            # now unselect the selected products to avoid them staying selected
            # after they get greyed out
            p.set_selected(False)

        self.update_button_sensitivity()

        if free_products:
            self.debug('Getting free products')
            downloaded = self.download_free_products(free_products)
            if not downloaded:
                self.debug('Not downloaded, leaving clicked handler.')
                return

        if pay_products:
            self.debug('Getting for pay products')
            self.get_products(pay_products)

        return

    def delete_event(self, widget, event, data=None):
        self.shutdown()
        return False

    def action_toggle_view(self, action, current):
        if self.actiongroup.get_action('ViewAllAvailable').get_active():
            self.listview.show_all_available()
            no_rows_txt = _('No plugins available for your system yet.')
        elif self.actiongroup.get_action('ViewMatching').get_active():
            self.listview.show_matching(self.required_plugins)
            no_rows_txt = _(
                'No matching plugins available for your system yet.')

        if self.listview.get_visible_rows() == 0:
            self.set_status_text(no_rows_txt)
            self.no_plugins_box.show_all()
        else:
            self.no_plugins_box.hide()

    def do_install(self, filenames):
        failed_installs = []
        for archive_fn in filenames:
            self.set_status_text(_("Installing '%s'") % archive_fn)
            while gtk.main_iteration_do(False):
                pass
            err_names = install.install_plugin_archive(archive_fn)
            if err_names:
                failed_installs.append(archive_fn)
            
        if failed_installs:
            dlg = gtk.MessageDialog(parent=self, flags=0,
                type=gtk.MESSAGE_ERROR,
                buttons=gtk.BUTTONS_CLOSE,
                message_format=_('Failed to install plugins'))
            dlg.set_title(_('Error'))
            txt = _('The following plugin archives could not be installed.')
            txt += ' '
            txt += _('Maybe they do not contain GStreamer plugins for this '
                    'architecture, or your GStreamer version is older than '
                    'the version required')
            txt += '\n'
            for f in failed_installs:
                txt += "\n%s" % f
            dlg.format_secondary_text(txt)
            dlg.run()
            dlg.destroy()
            self.set_status_text('')
        else:
            self.set_status_text(_('Installation successful'))

      # FIXME: we should only change the exit code if we installed some
      # of the _required_ plugins.
        if len(failed_installs) < len(filenames):
            self.installed_some = True
            self.exit_code = GST_INSTALL_PLUGINS_PARTIAL_SUCCESS

      # refresh display (listview in particular) to show new status
        self.queue_draw()
        return

    def action_install_archive(self, action):
        dlg = gtk.FileChooserDialog(
                  title=_('Select downloaded plugin archive'), parent=self,
                  buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                           gtk.STOCK_OK, gtk.RESPONSE_OK))
        desktopdir = os.path.join(os.path.expanduser('~'), 'Desktop')
        if os.access(desktopdir, os.F_OK):
            dlg.set_current_folder(desktopdir)
        dlg.set_select_multiple(True)
        dlg.set_default_response(gtk.RESPONSE_OK)
        ffilter_all = gtk.FileFilter()
        ffilter_all.set_name(_("All files"))
        ffilter_all.add_pattern("*")
        dlg.add_filter(ffilter_all)
        ffilter_tars = gtk.FileFilter()
        ffilter_tars.set_name(_("Plugin archives"))
        ffilter_tars.add_pattern("*.tar.gz")
        ffilter_tars.add_pattern("*.tar.bz2")
        dlg.add_filter(ffilter_tars)
        dlg.set_filter(ffilter_tars)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            filenames = dlg.get_filenames()
        else:
            filenames = None
        dlg.destroy()

        if filenames and len(filenames):
            self.do_install(filenames)
        return

    def action_quit(self, action):
        self.shutdown()
        return

    def action_about(self, action):
        gtk.about_dialog_set_url_hook(self.open_url)
        dlg = gtk.AboutDialog()
        dlg.set_name('Codeina')
        dlg.set_version(configure.VERSION)
        url = myutils.append_system_details_to_url('http://shop.fluendo.com')
        dlg.set_website(url)
        dlg.set_website_label('Fluendo codec shop')
        dlg.set_authors(["Tim-Philipp Müller <tim centricular net>"])
        dlg.set_copyright('Copyright © 2007 Fluendo S.A.')
        dlg.set_license(_('GNU General Public License\n'
                          'See http://www.gnu.org/copyleft/gpl.html '
                          'for more details'))
        dlg.set_comments(_('Fluendo Codec Installation Application'))
        dlg.connect('response', self.destroy_about_dialog)
        dlg.connect('close', self.destroy_about_dialog)
        dlg.run()
        return

    def destroy_about_dialog(self, dialog, data=None):
        dialog.destroy()
        return

    def open_url(self, dialog, link, data=None):
        myutils.open_url_in_browser(link)
        return

    def run(self):
        self.debug('Running main loop')
        gtk.main()
        self.debug('Ran main loop')
        return self.exit_code

    def shutdown(self):
        gtk.main_quit()
        return

    def update_button_sensitivity(self):
        # makes the "get" button clickable if there is at least one product
        # selected
        if self.product_list.count_selected():
            self.get_button.set_sensitive(True)
        else:
            self.get_button.set_sensitive(False)
 
    def set_status_text(self, txt):
        self.statusbar.pop(self.statusbar_ctx_id)
        if txt:
            self.statusbar.push(self.statusbar_ctx_id, txt)


def main(argv):
    gobject.threads_init()
    log.init('CODEINA_DEBUG')
    log.setPackageScrubList('codeina')

    check.initial_checks()

    # set default icon for window and about dialog
    icon_path = configure.make_icon_path('codeina.png')
    gtk.window_set_default_icon_from_file (icon_path)

    # create main application window
    app = CodeinaApp(argv)
    if app.show_distro_dialog() == GST_DISTRO_DIALOG_ABORT:
        return GST_INSTALL_PLUGINS_USER_ABORT

    app.show()
    try:
        exit_code = app.run()
    except KeyboardInterrupt:
        exit_code = GST_INSTALL_PLUGINS_USER_ABORT
    log.info('codeina', "bye (exit code = %d)" % exit_code)
    return exit_code
