# -*- Mode: Python -*-
# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Codeina, Codec Installation Application
# Copyright (C) 2007 Fluendo S.A.
#
# Written by Tim-Philipp Müller <tim centricular net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import os
import time

import check
import gst
from codeina import fxml

class ProductListParser(fxml.Parser):
    """
    I parse the product list.

    @rtype: L{ProductList}
    """
    def parse(self, handle, id=None):
        """
        Parse from the given file handle.
        The id is used to report problems.
        """
        root = self.getRoot(handle)
        node = root.documentElement

        if node.nodeName != 'products':
            print "No products"
            return

        return self._parse_products(node)

    # parser routines
    def _parse_products(self, node):
        products = {}
        def add_product(p):
            if p.nick in products.values():
              print "WARNING: more than one product with nick", p.nick
            else:
              products[p.nick] = p

        parsers = { 'product': (self._parse_product, add_product) }
        self.parseFromTable(node, parsers)

        return ProductList(products=products.values())

    def _parse_product(self, node):
        d = {}
        required = ('nick', 'id', 'logo-uri', 'price', 'provides', 'rank')
        optional = ('eula-uri', )
        attrs = self.parseAttributes(node, required=required, optional=optional)
        all = required + optional
        for i in range(0, len(all)):
            if all[i]:
                attr = all[i].replace('-', '_')
                d[attr] = attrs[i]
        names = []
        summaries = []
        flavors = []

        parsers = {
            'name': (self._parse_name, names.append),
            'summary': (self._parse_name, summaries.append),
            'flavor': (self._parse_flavor, flavors.append),
        }
        self.parseFromTable(node, parsers)

        # split the provides string into the individual parts
        provides = d['provides']
        d['provides'] = []
        for ppart in provides.split(';'):
                d['provides'].append(ppart.strip())

        # check price
        price_string = d['price']
        price = 0.0
        if price_string and price_string[:3] == "EUR":
            price = float(price_string[3:])
        d['price'] = price

        d['rank'] = int(d['rank'])
        d['id'] = int(d['id'])
        d['name'] = self._get_localized(names)
        d['summary'] = self._get_localized(summaries)
        d['flavors'] = flavors
        product = Product(**d)

        # Select the free-not-fully-installed plugins by default
        if price == 0.0:
            product.selected = not product.is_fully_installed()

        return product

    def _parse_flavor(self, node):
        d = {}
        required = ('system', 'arch', 'distro', 'distro-version',
            'core-req', 'base-req', 'last-modified', 'uri', 'comment')
        optional = ('sha1sum',)
        attrs = self.parseAttributes(node, required=required, optional=optional)
        for i in range(0, len(required)):
            # pesky - -> _ translation for attributes
            key = '_'.join(required[i].split('-'))
            d[key] = attrs[i]

        # create Version objects for the versions
        for k in 'core_req', 'base_req':
            d[k] = check.Version(d[k])

        # convert date string to time tuple for 'last-modified'
        d['last_modified'] = time.strptime(d['last_modified'], '%Y-%m-%d')

        flavor = ProductFlavor(**d)
        return flavor

    def _parse_name(self, node):
        # parse a name and its language
        # return tuple of language, name
        lang = 'C'
        if node.attributes:
            for a in node.attributes.items():
                if a[0] == u'xml:lang':
                    lang = a[1]

        return lang, node.childNodes[0].data

    def _get_localized(self, list, LL=None):
        # from a list of lang, str tuples, pick the one matching our locale
        if not LL:
            language = os.environ.get('LANGUAGE', None)
            if language != None:
                LL = language[:2]
            else:
                lang = os.environ.get('LANG', 'en')
                LL = lang[:2]

        # english is C for now
        if LL == 'en': LL = 'C'

        for lang, text in list:
            if lang == 'C':
                C = text
            if lang == LL:
                return text

        # we can always fall back to C
        return C

class XMLObject(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

class ProductList(XMLObject):
    products = None

    def dbg_print_part_products(self):
        for p in self.products:
            for p2 in self.products:
                if p.is_part_product_of(p2):
                    print "Product '%s' is part of product '%s'" % (p.name,
                            p2.name)

    def count_selected(self):
        num = 0
        for p in self.products:
            if p.selected:
                num += 1
        return num

    def get_selected(self):
        sel = []
        for p in self.products:
            if p.selected:
                sel.append(p)
        return sel


class Product(XMLObject):
    """
    Describes a product that can be bought.
    """
    name = None
    rank = 0
    price = 0
    logo_uri = None
    eula_uri = None
    provides = None
    flavors = None
    summary = None

    selected = False # for list view

    def __repr__(self):
        return "<%s for %s> " % (self.__class__.__name__, self.name)

    def get_selected(self):
        return self.selected

    def set_selected(self, desired_status):
        self.selected = desired_status

    def find_compatible_flavors(self, system, arch, distro, core_version,
                                base_version):
        ret = []
        for f in self.flavors:
            if f.is_compatible(system, arch, distro, core_version,
                               base_version):
                ret.append(f)
        return ret

    def get_flavors_for_current_system(self):
        return self.find_compatible_flavors(check.OS_NAME, check.ARCH_NAME,
                check.DISTRO_NAME, check.GST_CORE_VERSION, check.GST_PLUGINS_BASE_VERSION)

    def get_best_flavor_for_current_system(self):
        flavors = self.get_flavors_for_current_system()
        if not flavors:
            return None
        else:

      # FIXME: sort by required core version or something

            return flavors[0]

    def get_provided_elements(self):
        provided_elements = []
        for p in self.provides:
            if p[0:8] == "element-":
                provided_elements.append(p[8:])
        return provided_elements

    def get_provided_decoders(self):
        provided_decoders = []
        for p in self.provides:
            if p[0:8] == "decoder-":
                provided_decoders.append(p[8:])
        return provided_decoders

    def get_provided_encoders(self):
        provided_encoders = []
        for p in self.provides:
            if p[0:8] == "encoder-":
                provided_encoders.append(p[8:])
        return provided_encoders

    def get_provided_urisources(self):
        provided_sources = []
        for p in self.provides:
            if p[0:10] == "urisource-":
                provided_sources.append(p[10:])
        return provided_sources

    def get_provided_urisinks(self):
        provided_sinks = []
        for p in self.provides:
            if p[0:8] == "urisink-":
                provided_sinks.append(p[8:])
        return provided_sinks

  # returns (num_installed, num_missing)
  # checks install status of elements provided by this plugin against
  # the best available flavour for the current system/setup

    def get_installed_status(self):
        #flav = self.get_best_flavor_for_current_system()
        provided_elements = self.get_provided_elements()
        num_installed = 0
        num_missing = 0
        for element_name in provided_elements:
            version = check.get_element_version(element_name)
            if not version:
                num_missing += 1
            else:
                num_installed += 1
        return (num_installed, num_missing)

  # FIXME: maybe we should cache the installed status?

    def is_fully_installed(self):
        (num_inst, num_missing) = self.get_installed_status()
        return num_missing == 0

    def is_partially_installed(self):
        (num_inst, num_missing) = self.get_installed_status()
        return num_missing > 0 and num_inst > 0

    def updates_available(self):
        (num_inst, num_missing) = self.get_installed_status()
        if num_inst == 0:
            return False

      # make list of plugins to check (to minimise number of stat() calls)
        plugins = []
        highest_mtime = None
        for element_name in self.get_provided_elements():
            plugin = check.get_plugin_from_element_factory_name (element_name)
            if plugin and not plugin in plugins:
                plugins.append(plugin)

      # check last modification time for each plugin and find the highest
        for plugin in plugins:
            plugin_mtime = check.get_plugin_mtime(plugin)
            if not highest_mtime or plugin_mtime > highest_mtime:
                highest_mtime = plugin_mtime

        flav = self.get_best_flavor_for_current_system()
        if flav and highest_mtime and highest_mtime < flav.last_modified:
            return True
        else:
            return False

  # returns True if self provides all the element other_product provides and
  # other_product provides at least one element that self doesn't contain.

    def is_part_product_of(self, other_product):
        elements = self.get_provided_elements()
        other_elements = other_product.get_provided_elements()
        for e in elements:
            if not e in other_elements:
                return False

    # we don't just compare the lengths to guard against duplicate entries

        for e in other_elements:
            if not e in elements:
                return True
        return False

    def provides_element(self, element_name):
        elements = self.get_provided_elements()
        return elements and element_name in elements

    def provides_codec(self, provided_caps_list, wanted_caps_string):
        wanted_caps = gst.caps_from_string(wanted_caps_string)
        for provided_caps_string in provided_caps_list:
            provided_caps = gst.caps_from_string(provided_caps_string)
            if not provided_caps:
                print "Failed to convert string '%s' to GstCaps" % \
                    provided_caps_string
                continue
            intersect = provided_caps.intersect(wanted_caps)
            if intersect.get_size() > 0:
                return True
        return False

    def provides_decoder(self, caps_string):
        return self.provides_codec(self.get_provided_decoders(),
                                   caps_string)

    def provides_encoder(self, caps_string):
        return self.provides_codec(self.get_provided_encoders(),
                                   caps_string)

    def provides_urisource(self, protocol):
        sources = self.get_provided_urisources()
        return sources and protocol in sources

    def provides_urisink(self, protocol):
        sinks = self.get_provided_urisinks()
        return sinks and protocol in sinks

    def provides_requirement(self, required_plugin):
        if required_plugin and required_plugin.valid:
            if required_plugin.req_type == 'element':
                return self.provides_element(required_plugin.req_detail)
            if required_plugin.req_type == 'decoder':
                return self.provides_decoder(required_plugin.req_detail)
            if required_plugin.req_type == 'encoder':
                return self.provides_encoder(required_plugin.req_detail)
            if required_plugin.req_type == 'urisink':
                return self.provides_urisink(required_plugin.req_detail)
            if required_plugin.req_type == 'urisource':
                return self.provides_urisource(required_plugin.req_detail)
        print "Invalid required plugin object", required_plugin
        return False

class ProductFlavor(XMLObject):
    """
    Describes a flavor of a product.
    """
    system = None
    arch = None
    distro = None
    distro_version = None
    core_req = None
    base_req = None
    version = None
    uri = None
    comment = None
    sha1sum = None

    def is_compatible(self, system, arch, distro, core_version,
                      base_version):
        if system and system != self.system:
            return 0
        if arch and arch != self.arch:
            return 0

    # 'generic' fits any distro

        if distro and distro != self.distro and self.distro != 'generic':
            return 0
        if core_version and not self.core_req.fulfils_requirement(core_version):
            return 0
        if base_version and not self.base_req.fulfils_requirement(base_version):
            return 0
        return 1


if __name__ == "__main__":
    xml_data = open("../data/available-plugins.xml")
    parser = ProductListParser()
    plist = parser.parse(xml_data)
    print "done, %d products:" % len(plist.products)
    for p in plist.products:
        print " * %s (%.2f EUR)" % (p.name, p.price)
        for f in p.flavors:
            print "     - %s %s %s %s %s" % (f.system, f.arch, f.distro, os.path.basename(f.uri), f.sha1sum)

    f = plist.products[9].get_best_flavor_for_current_system()
    print "Best flavor:"
    print " - %s %s %s %s %s" % (f.system, f.arch, f.distro, os.path.basename(f.uri), f.sha1sum)

    flavs = plist.products[9].find_compatible_flavors ("Linux", "i386", None, check.Version("0.10.4"), check.Version("0.10.4"))
#    flavs = plist.products[0].find_compatible_flavors ("Linux", "ppc64", None, Version("0.10.9"), Version ("0.10.9"))
    print "Matching flavors:"
    for f in flavs:
        print " - %s %s %s %s %s" % (f.system, f.arch, f.distro, os.path.basename(f.uri), f.sha1sum)

